(*****************************************************************************
 * HOL-TestGen --- theorem-prover based test case generation
 *                 http://www.brucker.ch/projects/hol-testgen/
 *
 * Bank.thy --- an BirthdayBook inspired, elementary
 *              example for sequence testing based on a 
 *              deterministic specification.
 *
 * This file is part of HOL-TestGen.
 *
 * Copyright (c) 2009-15 Univ. Paris-Sud, France
 *               2009-15 Achim D. Brucker, Germany
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above
 *       copyright notice, this list of conditions and the following
 *       disclaimer in the documentation and/or other materials provided
 *       with the distribution.
 *
 *     * Neither the name of the copyright holders nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************)

theory  MyKeOS_test_conc
imports MyKeOS (* includes Testing *) 
        (*"../../../src/codegen_fsharp/Code_Integer_Fsharp"*) 
        "~~/src/HOL/Library/Code_Target_Numeral"
        "../../../src/codegen_gdb/Code_gdb_script"
begin

declare [[testgen_profiling]]

section{* Interleaving *}
text{* The purpose of this example is to model system calls that consists of a number of
   (internal) atomic actions; the global behavior is presented by the interleaving of 
   the actions actions *}

definition "syscall tid thid m m' = [alloc tid thid m, release tid thid m', status tid thid]"

value "interleave (syscall 5 0 m m') (syscall 5 1 m m')"

text{* In the following, we do a predicate abstraction on the interleace language, leading to
an automaton represented as a set of rewrites ... *}
fun Interleave :: "in_c list \<Rightarrow> nat \<times> nat \<Rightarrow> int \<Rightarrow> nat \<Rightarrow> nat \<Rightarrow> nat \<Rightarrow> nat \<Rightarrow> bool" (infixl "\<bowtie>"100)
where    "S \<bowtie> (a, b) = (\<lambda> tid m m' m'' m'''. (S \<in> interleave (drop a (syscall tid 0 m m'))
                                                              (drop b (syscall tid 1 m'' m'''))))"
lemma init_Interleave : "(S \<bowtie> (0, 0)) tid m m' m'' m''' = 
                                       (S \<in> interleave (syscall tid 0 m m') 
                                                       (syscall tid 1 m'' m'''))"
by simp

value "interleave (syscall tid 0 m m') (syscall tid 1 m'' m''')"

find_theorems name:"Interleave"
(* refusals *)
lemma ref_mt [simp]: "\<not>([] \<bowtie> (0, 0)) tid m m' m'' m'''"
by (simp add: syscall_def)
lemma ref_0_0 [simp]: "\<not>(((status a b) # R) \<bowtie> (0, 0)) tid m m' m'' m'''"
by (simp add: syscall_def)
lemma ref_1_0 [simp]: "\<not>(((status a b) # R) \<bowtie> (1, 0)) tid m m' m'' m'''"
by (simp add: syscall_def)
lemma ref_0_1 [simp]: "\<not>(((status a b) # R) \<bowtie> (0, 1)) tid m m' m'' m'''"
by (simp add: syscall_def)
lemma ref_1_1 [simp]: "\<not>(((status a b) # R) \<bowtie> (1, 1)) tid m m' m'' m'''"
by (simp add: syscall_def)
lemma ref_3_1 [simp]: "\<not>(((status a b) # R) \<bowtie> (3, 1)) tid m m' m'' m'''"
by (simp add: syscall_def)
lemma ref_1_3 [simp]: "\<not>(((status a b) # R) \<bowtie> (1, 3)) tid m m' m'' m'''"
by (simp add: syscall_def)

(* transducer rules -- helps npothing for the moment. Is an experiment to
   be re-conciled with gen_test_cases. *)
lemma trans_0_0 [simp]: "(((a # R) \<bowtie> (0, 0)) tid m m' m'' m''') = 
                     ((a = alloc tid 0 m \<and> (R \<bowtie> (1, 0)) tid m m' m'' m''') \<or>
                      (a = alloc tid 1 m'' \<and> (R \<bowtie> (0, 1)) tid m m' m'' m'''))"
by (simp add: syscall_def, rule iffI, metis, metis)
lemma trans_1_0 [simp]: "(((a # R) \<bowtie> (1, 0)) tid m m' m'' m''') = 
                     ((a = release tid 0 m' \<and> (R \<bowtie> (2, 0)) tid m m' m'' m''') \<or>
                      (a = alloc tid 1 m'' \<and> (R \<bowtie> (1, 1)) tid m m' m'' m'''))"
by (simp add: syscall_def, rule iffI, metis, metis)
lemma trans_2_0 [simp]: "(((a # R) \<bowtie> (2, 0)) tid m m' m'' m''') = 
                        ( (a = status tid 0 \<and> R = [alloc tid 1 m'', release tid 1 m''', status tid 1]) \<or> 
                          (a = alloc tid 1 m'' \<and> (R \<bowtie> (2, 1)) tid m m' m'' m'''))"
by (simp add: syscall_def, rule iffI, metis, metis)

lemma trans_2_1 [simp]: "(((a # R) \<bowtie> (2, 1)) tid m m' m'' m''') = 
                        ( (a = status tid 0 \<and> R = [release tid 1 m''', status tid 1]) \<or> 
                          (a = release tid 1 m''' \<and> (R \<bowtie> (2, 2)) tid m m' m'' m'''))"
by (simp add: syscall_def, rule iffI, metis, metis)

lemma trans_2_2 [simp]: "(((a # R) \<bowtie> (2, 2)) tid m m' m'' m''') = 
                        ( (a = status tid 0 \<and> R = [status tid 1]) \<or> 
                          (a = status tid 1 \<and> R = [status tid 0]))"
by (simp add: syscall_def, rule iffI, metis, metis)

value "interleave (drop 0 (syscall tid 0 m m'))(drop 0 (syscall tid 1 m'' m'''))"

text{* TestData Hack:*} 

lemma PO_norm0 [simp]: "PO True" by(simp add: PO_def)


text{* The following scenario is meant to describe the symbolic execution step by step. *}

declare Monads.mbind'_bind [simp del]  (* stop eliminating elimination of @{term "mbind\<^sub>F\<^sub>a\<^sub>i\<^sub>l\<^sub>S\<^sub>t\<^sub>o\<^sub>p"} *)

find_theorems "mbind\<^sub>F\<^sub>a\<^sub>i\<^sub>l\<^sub>S\<^sub>t\<^sub>o\<^sub>p []"

lemma example_symbolic_execution_simulation :
   assumes H:  "S = [alloc tid 1 m'', release tid 0 m', release tid 1 m''', status tid 1]"
   assumes SE: "\<sigma>\<^sub>0 \<Turnstile> (s \<leftarrow> mbind\<^sub>F\<^sub>a\<^sub>i\<^sub>l\<^sub>S\<^sub>t\<^sub>o\<^sub>p S SYS; return (x = s))"
   shows       "P"
apply(insert SE H)
apply(hypsubst)
apply(tactic "ematch_tac [@{thm status.exec_mbindFStop_E}, @{thm release.exec_mbindFStop_E},
                          @{thm alloc.exec_mbindFStop_E}] 1")
apply(tactic "ematch_tac [@{thm status.exec_mbindFStop_E}, @{thm release.exec_mbindFStop_E},
                          @{thm alloc.exec_mbindFStop_E}] 1")
apply(tactic "ematch_tac [@{thm status.exec_mbindFStop_E}, @{thm release.exec_mbindFStop_E},
                          @{thm alloc.exec_mbindFStop_E}] 1")
apply(tactic "ematch_tac [@{thm status.exec_mbindFStop_E}, @{thm release.exec_mbindFStop_E},
                          @{thm alloc.exec_mbindFStop_E}] 1")
apply(tactic "ematch_tac [@{thm status.exec_mbindFStop_E}, @{thm release.exec_mbindFStop_E},
                          @{thm alloc.exec_mbindFStop_E},  @{thm valid_mbind'_mtE} ] 1")
apply simp
oops

lemma 
 assumes valid:  "(\<sigma> \<Turnstile> ( s \<leftarrow> mbind\<^sub>F\<^sub>a\<^sub>i\<^sub>l\<^sub>S\<^sub>a\<^sub>v\<^sub>e (alloc c no m # S) SYS; unit\<^sub>S\<^sub>E (P s)))"
 and case1:
       "(c, no) \<in> dom \<sigma> \<Longrightarrow>
         \<sigma>((c, no) \<mapsto> the (\<sigma> (c, no)) + int m) \<Turnstile>
             (s \<leftarrow> mbind\<^sub>F\<^sub>a\<^sub>i\<^sub>l\<^sub>S\<^sub>a\<^sub>v\<^sub>e S SYS; unit\<^sub>S\<^sub>E (P (alloc_ok # s))) \<Longrightarrow> Q"
 and case2: "(c, no) \<notin> dom \<sigma> \<Longrightarrow> \<sigma> \<Turnstile> unit\<^sub>S\<^sub>E (P []) \<Longrightarrow> Q"
 shows "Q"
 apply (insert assms)
 apply (erule MyKeOS.alloc.exec_mbindFSave_E)
apply metis
by metis
(*by (metis (mono_tags) alloc.exec_mbindFSave_If)*)
thm MyKeOS.alloc.exec_mbindFSave_E

ML{*hyp_subst_tac *}
test_spec test_status:
assumes account_defined: "(tid,0) \<in> dom \<sigma>\<^sub>0 \<and> (tid,1) \<in> dom \<sigma>\<^sub>0" 
and     test_purpose   : "S \<in> interleave (syscall tid 0 m m') (syscall tid 1 m'' m''')"
and     sym_exec_spec  :
       " \<sigma>\<^sub>0 \<Turnstile> (s \<leftarrow> mbind\<^sub>F\<^sub>a\<^sub>i\<^sub>l\<^sub>S\<^sub>t\<^sub>o\<^sub>p S SYS; return (x = s))"
shows  " \<sigma>\<^sub>0 \<Turnstile> (s \<leftarrow> mbind\<^sub>F\<^sub>a\<^sub>i\<^sub>l\<^sub>S\<^sub>t\<^sub>o\<^sub>p S PUT; return (s = x))"  
apply(insert (* store_finite *)  account_defined test_purpose sym_exec_spec)
apply(frule length_interleave)
(* direct enuleration of the interleaves *)
apply(simp add: syscall_def)
apply(tactic "TestGen.mp_fy 1",rule_tac x=x in spec[OF allI])
txt{* just case elaboration of test-cases *}
apply (clarify, elim disjE)
apply (tactic "ALLGOALS(hyp_subst_tac @{context})")
txt{* symbolic execution *}
apply(tactic "ALLGOALS(TestGen.REPEAT'(ematch_tac [@{thm status.exec_mbindFStop_E},
                                                   @{thm release.exec_mbindFStop_E},
                                                   @{thm alloc.exec_mbindFStop_E},
                                                   @{thm valid_mbind'_mtE}
                          ]))")

txt{* elimination of infeasible executions *}
apply(simp_all)
apply(tactic "ALLGOALS(hyp_subst_tac @{context})")
apply(tactic "ALLGOALS(TestGen.uniformityI_tac  @{context} [\"PUT\"])")
mk_test_suite "mykeos_interleave"

(* Use the SMT backend *)
declare [[testgen_iterations=0]]
declare [[testgen_SMT]]
gen_test_data "mykeos_interleave"
thm mykeos_interleave.concrete_tests

generate_test_script "mykeos_interleave"
thm                   mykeos_interleave.test_script

subsection{*Generation of a gdb file*}

ML {*   (*building the gdb term*)
       fun actTerm_to_gdbTerm (Const(@{const_name "alloc"}, typ) $ _ $ B $ _)= 
        gdb_break_point_entry (Const(@{const_name "alloc"}, typ)) B $ 
        gdb_break_point_exist @{term "0"} B 
      | actTerm_to_gdbTerm (Const(@{const_name "release"}, typ) $ _ $ B $ _)= 
        gdb_break_point_entry (Const(@{const_name "release"}, typ)) B $
        gdb_break_point_exist @{term "0"} B
      | actTerm_to_gdbTerm (Const(@{const_name "status"}, typ) $ _ $ B)= 
        gdb_break_point_entry (Const(@{const_name "status"}, typ)) B $
        gdb_break_point_exist @{term "0"} B
      | actTerm_to_gdbTerm (Const(@{const_name "end"}, _) $ _)= 
        gdb_start_term $ gdb_endFile;
 *}

ML{*   type info_threads = {task_id: int, 
                            th_id: int, 
                            order: int, 
                            break_alloc: int * int,
                            break_release: int * int,
                            break_status: int * int,
                            break_main: int * int}; 
   *}

(* Remark bu: from the model (i.e.) the list of mykeos_interleave.test_data, one could extract
   automatically given the following interface *)
ML{*  type info_threads_configure = {input_type : typ,
                                     get_task_id : term -> int,
                                                   (* precond: term must be of type typ*)
                                     get_thread_id : term -> int,
                                                   (* precond: term must be of type typ*)
                                     config_atomic_actions : (string ->  int * int) -> unit,
                                     set_break_main : (int * int) -> unit
                                    };

(* So, the package gdb_script_generator can provide a function: *)

fun generate_gdb_script_config (X: info_threads_configure) 
                               (testenv : string) (* in this case: mykeos_interleave *)
    = []: (string * int * int) list  


*}


ML {* @{thms  mykeos_interleave.concrete_tests}*}
ML{*
 
val thread_info1 = {task_id = 5 , th_id = 1 , order = 2,
                    break_alloc = (50, 59), break_release= (59, 61), break_status =  (61, 63),
                    break_main = (123, 137)}; 

val thread_info12 = {task_id = 5 , th_id = 0 , order = 3,
                     break_alloc = (68, 77), break_release= (77, 79), break_status =  (79, 81),
                     break_main = (123, 137)}; 

val thread_info2 = {task_id = 3 , th_id = 1 , order = 2,
                    break_alloc = (94 , 96), break_release= (96, 98), break_status =  (98, 100),
                    break_main = (123, 137)};
  
val thread_info31 = {task_id = 3 , th_id = 0 , order = 3,
                     break_alloc = (112, 114), break_release= (114, 116), break_status =  (116, 118),
                     break_main = (123, 137)}; 


val needed_informations = [thread_info1, thread_info12, thread_info2, thread_info31, 
                           thread_info2, thread_info31]

*}


ML {*
  
  fun check_identity  (info:info_threads) task_id th_id =
       ((info |> #task_id) = (task_id |> HOLogic.dest_number|> snd) andalso
       (info |> #th_id) = (th_id |> HOLogic.dest_number|> snd));

  fun get_successor_order [] _ = 99 |> mk_number @{typ int}
  
  |   get_successor_order  ((info:info_threads)::infos) 
                           (Const(@{const_name "alloc"}, typ) $ task_id $ th_id $ value) =
       if check_identity info task_id th_id
       then info |> #order |> mk_number @{typ int} 
       else get_successor_order infos (Const(@{const_name "alloc"}, typ) $ task_id $ th_id $ value)
  |   get_successor_order  ((info:info_threads)::infos) 
                           (Const(@{const_name "release"}, typ) $ task_id $ th_id $ value) =
       if check_identity info task_id th_id
       then info |> #order |> mk_number @{typ int} 
       else get_successor_order infos (Const(@{const_name "release"}, typ) $ task_id $ th_id $ value)
   |  get_successor_order  ((info:info_threads)::infos) 
                           (Const(@{const_name "status"}, typ) $ task_id $ th_id) =
       if check_identity info task_id th_id
       then info |> #order |> mk_number @{typ int} 
       else get_successor_order infos (Const(@{const_name "status"}, typ) $ task_id $ th_id)
   | get_successor_order _ (_ $ _ $ _ $ _ $ _ $ _ $ _ $ _ $ _ $ _ $ term11 $ _ $ _ $ _ $ _ $ _ $ _) 
                           = term11;

  fun inputSeq_to_gdbEn [] [] = []
  |   inputSeq_to_gdbEn _ [] = []
  |   inputSeq_to_gdbEn [] terms = terms

  |   inputSeq_to_gdbEn ((info:info_threads)::infos)
                       ([(Const(@{const_name "alloc"}, typ) $ task_id $ th_id $ value)]) = 
      if check_identity info task_id th_id 
      then gdb_break_point_entry (info |> #break_alloc |> fst |> mk_number @{typ int}) 
                                 (info |> #order |> mk_number @{typ int}) :: []
      else inputSeq_to_gdbEn infos [(Const(@{const_name "alloc"}, typ) $ task_id $ th_id $ value)]

  |   inputSeq_to_gdbEn ((info:info_threads)::infos)
                       ([(Const(@{const_name "release"}, typ) $ task_id $ th_id $ value)]) = 
      if check_identity info task_id th_id 
      then gdb_break_point_entry (info |> #break_release |> fst |>mk_number @{typ int}) 
                                 (info |> #order |> mk_number @{typ int}):: []
      else inputSeq_to_gdbEn infos [(Const(@{const_name "release"}, typ) $ task_id $ th_id $ value)]

  |   inputSeq_to_gdbEn ((info:info_threads)::infos)
                       ([(Const(@{const_name "status"}, typ) $ task_id $ th_id)]) = 
      if check_identity info task_id th_id 
      then gdb_break_point_entry (info |> #break_status |> fst |> mk_number @{typ int}) 
                                 (info |> #order |> mk_number @{typ int}) :: []
      else inputSeq_to_gdbEn infos [(Const(@{const_name "status"}, typ) $ task_id $ th_id)]

  |   inputSeq_to_gdbEn ((info:info_threads)::infos)
                        ((Const(@{const_name "alloc"}, typ) $ task_id $ th_id $ value)::terms) = 
      if check_identity info task_id th_id  
      then(gdb_break_point_entry (info |> #break_alloc |> fst |> mk_number @{typ int}) 
                                 (info |> #order |> mk_number @{typ int})):: 
                 inputSeq_to_gdbEn (info::infos) terms
      else inputSeq_to_gdbEn infos
          ((Const(@{const_name "alloc"}, typ) $ task_id $ th_id $ value)::
           inputSeq_to_gdbEn (info::infos) terms)

   |  inputSeq_to_gdbEn ((info:info_threads)::infos)
                       ((Const(@{const_name "release"}, typ) $ task_id $ th_id $ value)::terms) = 
      if check_identity info task_id th_id  
      then (gdb_break_point_entry (info |> #break_release |> fst |> mk_number @{typ int}) 
                                 (info |> #order |> mk_number @{typ int})):: 
                 inputSeq_to_gdbEn (info::infos) terms
      else inputSeq_to_gdbEn infos
          ((Const(@{const_name "release"}, typ) $ task_id $ th_id $ value)::
           inputSeq_to_gdbEn (info::infos) terms)

   |  inputSeq_to_gdbEn ((info:info_threads)::infos)
                      ((Const(@{const_name "status"}, typ) $ task_id $ th_id)::terms) = 
     if check_identity info task_id th_id  
     then (gdb_break_point_entry (info |> #break_status |> fst |> mk_number @{typ int}) 
                                (info |> #order |> mk_number @{typ int})) :: 
                inputSeq_to_gdbEn (info::infos) terms
     else inputSeq_to_gdbEn infos
         ((Const(@{const_name "status"}, typ) $ task_id $ th_id)::
          inputSeq_to_gdbEn (info::infos) terms)
   |inputSeq_to_gdbEn infos (term1 $ term2 $ term3 $ term4 $ term5 $ term6 $ ter7 $ term8 $ 
                             term9 $ term10 $ term11 $ term12 $ term13 $ term14 $ term15 $ term16 $ 
                             term17::terms) = 
     term1 $ term2 $ term3 $ term4 $ term5 $ term6 $ ter7 $ term8 $ 
     term9 $ term10 $ term11 $ term12 $ term13 $ term14 $ term15 $ term16 $ 
                               term17  :: inputSeq_to_gdbEn infos terms ;

   (*|  inputSeq_to_gdbEn infos (term::terms) = 
      term :: inputSeq_to_gdbEn infos terms*)


  fun inputSeq_to_gdbEx [] [] = []
  |   inputSeq_to_gdbEx _ [] = []
  |   inputSeq_to_gdbEx [] terms = terms

  |   inputSeq_to_gdbEx ((info:info_threads)::infos)
                       ([(Const(@{const_name "alloc"}, typ) $ task_id $ th_id $ value)]) = 
      if check_identity info task_id th_id 
      then gdb_break_point_exist (info |> #break_alloc |> snd |> mk_number @{typ int}) 
            (info |> #order |> mk_number @{typ int}):: []

      else inputSeq_to_gdbEx infos [(Const(@{const_name "alloc"}, typ) $ task_id $ th_id $ value)]

  |   inputSeq_to_gdbEx ((info:info_threads)::infos)
                       ([(Const(@{const_name "release"}, typ) $ task_id $ th_id $ value)]) = 
      if check_identity info task_id th_id 
      then gdb_break_point_exist (info |>  #break_release |> snd |> mk_number @{typ int}) 
            (info |> #order |> mk_number @{typ int}):: []
      else inputSeq_to_gdbEx infos [(Const(@{const_name "release"}, typ) $ task_id $ th_id $ value)]

  |   inputSeq_to_gdbEx ((info:info_threads)::infos)
                       ([(Const(@{const_name "status"}, typ) $ task_id $ th_id)]) = 
      if check_identity info task_id th_id 
      then gdb_break_point_exist (info |> #break_status |> snd |> mk_number @{typ int}) 
           (info |> #order |> mk_number @{typ int}):: []
      else inputSeq_to_gdbEx infos [(Const(@{const_name "status"}, typ) $ task_id $ th_id )]

  |   inputSeq_to_gdbEx ((info:info_threads)::infos)
                        ((Const(@{const_name "alloc"}, typ) $ task_id $ th_id $ value)::terms) = 
      if check_identity info task_id th_id  
      then gdb_break_point_exist (info |> #break_alloc |> snd |> mk_number @{typ int}) 
                                 (if terms = [] 
                                  then (info |> #order |> mk_number @{typ int})  
                                  else get_successor_order (info::infos) 
                                                           (hd terms)):: 
                 inputSeq_to_gdbEx (info::infos) terms
      else inputSeq_to_gdbEx infos
          ((Const(@{const_name "alloc"}, typ) $ task_id $ th_id $ value)::
           terms)

   |  inputSeq_to_gdbEx ((info:info_threads)::infos)
                       ((Const(@{const_name "release"}, typ) $ task_id $ th_id $ value)::terms) = 
      if check_identity info task_id th_id  
      then gdb_break_point_exist (info |> #break_release |> snd |> mk_number @{typ int}) 
                                 (if terms = [] 
                                  then (info |> #order |> mk_number @{typ int})  
                                  else get_successor_order (info::infos) 
                                                           (hd terms)) :: 
                 inputSeq_to_gdbEx (info::infos) terms
      else inputSeq_to_gdbEx infos
          ((Const(@{const_name "release"}, typ) $ task_id $ th_id $ value)::
            terms)

   |  inputSeq_to_gdbEx ((info:info_threads)::infos)
                      ((Const(@{const_name "status"}, typ) $ task_id $ th_id)::terms) = 
     if check_identity info task_id th_id  
     then gdb_break_point_exist (info |> #break_status |> snd |> mk_number @{typ int}) 
                                (if terms = [] 
                                 then (info |> #order |> mk_number @{typ int})
                                 else get_successor_order (info::infos) 
                                                          (hd terms)) :: 
                inputSeq_to_gdbEx (info::infos) terms
     else inputSeq_to_gdbEx infos
         ((Const(@{const_name "status"}, typ) $ task_id $ th_id)::
           terms)
   
   |inputSeq_to_gdbEx infos (term1 $ term2 $ term3 $ term4 $ term5 $ term6 $ ter7 $ term8 $ 
                             term9 $ term10 $ term11 $ term12 $ term13 $ term14 $ term15 $ term16 $ 
                             term17::terms) = 
     term1 $ term2 $ term3 $ term4 $ term5 $ term6 $ ter7 $ term8 $ 
     term9 $ term10 $ term11 $ term12 $ term13 $ term14 $ term15 $ term16 $ 
                               term17  :: inputSeq_to_gdbEx infos terms 
    (*|  inputSeq_to_gdbEx infos (term::terms) = 
      term :: inputSeq_to_gdbEn infos terms*);
  gdb_break_main_entry ("main" |> HOLogic.mk_string);

  fun 
   add_gdb_main [] terms = terms
  |   add_gdb_main ((info:info_threads)::infos) 
                   (Const(@{const_name "alloc"}, typ) $ task_id $ th_id $ value::terms) = 
      if check_identity info task_id th_id 
      then 
           [gdb_break_main_entry (info |> #break_main |> fst |> mk_number @{typ int})$
            gdb_break_main_exit (info |> #break_main |> snd |> mk_number @{typ int}) 
                                  (info |> #order |> mk_number @{typ int})] 
      else add_gdb_main infos (Const(@{const_name "alloc"}, typ) $ task_id $ th_id $ value::terms)

  |   add_gdb_main ((info:info_threads)::infos) 
                   (Const(@{const_name "release"}, typ) $ task_id $ th_id $ value::terms) = 
      if check_identity info task_id th_id 
      then 
           [gdb_break_main_entry (info |> #break_main |> fst |> mk_number @{typ int})$
            gdb_break_main_exit (info |> #break_main |> snd |> mk_number @{typ int}) 
                                  (info |> #order |> mk_number @{typ int})] 
      else add_gdb_main infos (Const(@{const_name "release"}, typ) $ task_id $ th_id $ value::terms)

  |   add_gdb_main ((info:info_threads)::infos) 
                   (Const(@{const_name "status"}, typ) $ task_id $ th_id::terms) = 
      if check_identity info task_id th_id 
      then 
           [gdb_break_main_entry (info |> #break_main |> fst |> mk_number @{typ int})$
            gdb_break_main_exit (info |> #break_main |> snd |> mk_number @{typ int}) 
                                  (info |> #order |> mk_number @{typ int})] 
      else add_gdb_main infos (Const(@{const_name "status"}, typ) $ task_id $ th_id::terms)

  |  add_gdb_main _ _ = [];

*} 

ML {* 
    
gen_gdb_scripts' 
        inputSeq_to_gdbEn inputSeq_to_gdbEx add_gdb_main needed_informations 
       @{theory} "impl/c_conc/MyKeOS" (@{thms mykeos_interleave.concrete_tests});

    *}

section{* Experimental Space *}


declare[[testgen_trace]] 
(*
subsection{*Code Generation Setup*}

subsubsection{*Generation of an SML file to put datatypes*}

definition program_dum ::"int \<Rightarrow>  in_c \<Rightarrow> out_c \<Rightarrow>int "
where     "program_dum x a outs = 0"

export_code program_dum in SML
  module_name Datatypes file "impl/c_seq/datatypes1.sml"


subsubsection{*Code Setup for Datatypes*}
 
code_printing
  type_constructor in_c => (SML) "Datatypes.in'_c"
    |constant alloc   => (SML) "!(Datatypes.Alloc ( _ , _ , _))"
    |constant release => (SML) "!(Datatypes.Release ( _ , _ , _ ))"
    |constant status  => (SML) "!(Datatypes.Status ( _ , _ ))"

code_printing
  type_constructor out_c => (SML) "Datatypes.out'_c"
    |constant alloc_ok   => (SML) "Datatypes.Alloc'_ok"
    |constant release_ok => (SML) "Datatypes.Release'_ok "
    |constant status_ok  => (SML) "!(Datatypes.Status'_ok ( _ ))"

code_printing 
  type_constructor int => 
    (SML) "Datatypes.int"
    | constant int_of_integer =>
      (SML) "!(Datatypes.Int'_of'_integer ( _ ))"


code_printing 
  type_constructor nat => 
    (SML) "Datatypes.nat"
    | constant Nat => (SML) "!(Datatypes.Nat ( _ ))"


subsection{*HOL to SML adapter*}
subsubsection{*Constant definitions: stubs*}

consts MyKeOS_stub1:: "int \<Rightarrow> int \<Rightarrow> int \<Rightarrow>(int, '\<sigma>)MON\<^sub>S\<^sub>E"  

consts MyKeOS_stub2:: "int \<Rightarrow> int \<Rightarrow> int \<Rightarrow> (int, '\<sigma>)MON\<^sub>S\<^sub>E"

consts MyKeOS_stub3:: "int \<Rightarrow> int \<Rightarrow> int \<Rightarrow> (int, '\<sigma>)MON\<^sub>S\<^sub>E"

subsubsection{*Conversion: Integer to Action Output*}

(* hack : codegenerator does not work for "nat" now ! Therefore an own conversion function:*)
fun    my_nat_conv :: "int \<Rightarrow> nat"
where "my_nat_conv x =(if x <= 0 then 0 else Suc (my_nat_conv(x - 1)))"   


fun stubs_to_out::"in_c \<Rightarrow> int \<Rightarrow>  out_c"
where 
  "stubs_to_out (alloc task_id th_id res)   y = (if (y = (plus (int res) 1)) 
                                                  then alloc_ok else release_ok)"
| "stubs_to_out (release task_id th_id res)  y = (if (y = (minus (int res) 1)) 
                                                   then release_ok else alloc_ok)"
| "stubs_to_out (status task_id th_id)   y = (if (y = y) 
                                               then status_ok (my_nat_conv y) else release_ok)"
(*Problem with the state x it should be a function not an int*)
fun mykeAdapter::"in_c \<Rightarrow> (int \<times> int \<Rightarrow> int option) \<Rightarrow> (out_c \<times> (int \<times> int \<Rightarrow> int option)) option"
where 
  "mykeAdapter (alloc task_id th_id res)  x = (out \<leftarrow> MyKeOS_stub1 task_id th_id res; 
                               return(stubs_to_out (alloc task_id th_id res) 
                                                   ((fst o the) (MyKeOS_stub1 task_id th_id res x)))) x"
| "mykeAdapter (release task_id th_id res)  x = (out \<leftarrow> MyKeOS_stub2 task_id th_id res; 
                               return(stubs_to_out (release task_id th_id res) 
                                                   ((fst o the) (MyKeOS_stub2 task_id th_id res x)))) x"
| "mykeAdapter (status task_id th_id)  x = (out \<leftarrow> MyKeOS_stub3 task_id th_id (the (x (task_id, th_id))); 
                               return(status_ok ((my_nat_conv o the) (x (task_id, th_id))))) x"
term "stubs_to_out (release task_id th_id res) 
                   ((fst o the) (MyKeOS_stub2 task_id th_id res x))"

subsection{*Serialisation: semantics of stubs*}

code_printing
  constant MyKeOS_stub1 => (SML) "MyKeOSAdapter1.alloc"
code_printing
  constant MyKeOS_stub2 => (SML) "MyKeOSAdapter1.release"
code_printing
  constant MyKeOS_stub3 => (SML) "MyKeOSAdapter1.status"

(*HOL adapter in SML*)

export_code   mykeAdapter in SML
  module_name MykeAdapter file "impl/c_seq/mykeAdapter.sml"

subsection{*Serialisation: semantics of SUT*}

code_printing 
  constant PUT => (SML) "!(MykeAdapter.mykeAdapter ( _ ) ( _ ))"

export_code                  mykeos_simple.test_script in SML    
module_name TestScript file  "impl/c_seq/mykeos_test_script1.sml"*)
 
subsection{*Code Generation Setup For Concurrent Scenario*}

  
subsubsection{*Generation of an SML file to put datatypes*}

definition program_dum_conc 
           ::"(int \<times> int \<Rightarrow> int option) \<Rightarrow>  in_c \<Rightarrow> out_c \<Rightarrow> (int \<times> int \<Rightarrow> int option) "
where     "program_dum_conc \<sigma> a outs = [(0, 0)\<mapsto> 0]"

export_code program_dum_conc in SML
  module_name Datatypes file "impl/c_conc/datatypes.sml"

subsubsection{*Code Setup for Datatypes*}
 
code_printing
  type_constructor in_c => (SML) "Datatypes.in'_c"
    |constant alloc   => (SML) "!(Datatypes.Alloc ( _ , _ , _))"
    |constant release => (SML) "!(Datatypes.Release ( _ , _ , _ ))"
    |constant status  => (SML) "!(Datatypes.Status ( _ , _ ))"

code_printing
  type_constructor out_c => (SML) "Datatypes.out'_c"
    |constant alloc_ok   => (SML) "Datatypes.Alloc'_ok"
    |constant release_ok => (SML) "Datatypes.Release'_ok "
    |constant status_ok  => (SML) "!(Datatypes.Status'_ok ( _ ))"

code_printing 
  type_constructor int => 
    (SML) "Datatypes.int"
    | constant int_of_integer =>
      (SML) "!(Datatypes.Int'_of'_integer ( _ ))"


code_printing 
  type_constructor nat => 
    (SML) "Datatypes.nat"
    | constant Nat => (SML) "!(Datatypes.Nat ( _ ))"

subsection{*HOL to SML adapter*}

subsubsection{*Constant definitions: stubs*}

consts MyKeOS_conc1:: "int \<Rightarrow> int \<Rightarrow> int \<Rightarrow> (int, '\<sigma>)MON\<^sub>S\<^sub>E"  


subsubsection{*Conversion: Integer to Action Output*}

(* hack : codegenerator does not work for "nat" now ! Therefore an own conversion function:*)
fun    my_nat_conv :: "int \<Rightarrow> nat"
where "my_nat_conv x =(if x <= 0 then 0 else Suc (my_nat_conv(x - 1)))"   


fun stubs_to_out_conc::"in_c \<Rightarrow> (int \<times> int \<Rightarrow> int option) \<Rightarrow> int \<Rightarrow> out_c"
where 
  "stubs_to_out_conc (alloc task_id th_id res) \<sigma>   \<sigma>_impl = 
  (if (\<sigma>_impl = (plus ((the o \<sigma>)(task_id,th_id)) (int res))) 
   then alloc_ok 
   else release_ok)"
| "stubs_to_out_conc (release task_id th_id res) \<sigma>  \<sigma>_impl = 
   (if (\<sigma>_impl = (minus ((the o \<sigma>)(task_id,th_id)) (int res))) 
    then release_ok 
    else alloc_ok)"
| "stubs_to_out_conc (status task_id th_id) \<sigma>   \<sigma>_impl =
   (if (\<sigma>_impl = ((the o \<sigma>)(task_id,th_id))) 
    then status_ok (my_nat_conv \<sigma>_impl) 
    else release_ok)"




fun mykeAdapter_con::"in_c \<Rightarrow> (int \<times> int \<Rightarrow> int option) \<Rightarrow> (out_c \<times> (int \<times> int \<Rightarrow> int option)) option"
where 
  "mykeAdapter_con (alloc task_id th_id res)  \<sigma> = 
                   (out \<leftarrow> MyKeOS_conc1 task_id th_id res; 
                           return(stubs_to_out_conc (alloc task_id th_id res) \<sigma> 
                                               ((fst o the) (MyKeOS_conc1 task_id th_id res \<sigma>)))) \<sigma>"
| "mykeAdapter_con (release task_id th_id res)  \<sigma> = 
                   (out \<leftarrow> MyKeOS_conc1 task_id th_id res; 
                           return(stubs_to_out_conc (alloc task_id th_id res) \<sigma> 
                                               ((fst o the) (MyKeOS_conc1 task_id th_id res \<sigma>)))) \<sigma>"
| "mykeAdapter_con (status task_id th_id)  \<sigma> = 
                   (out \<leftarrow> MyKeOS_conc1 task_id th_id (the (\<sigma> (task_id, th_id))); 
                           return(status_ok ((my_nat_conv o the) (\<sigma> (task_id, th_id))))) \<sigma>"

subsection{*Serialisation: semantics of conc stubs*}

code_printing
  constant MyKeOS_conc1 => (SML) "!(MyKeOSAdapter.get'_state ( _ ) ( _ ) ( _ ) ( _ ))"

export_code   mykeAdapter_con in SML
  module_name MykeAdapter file "impl/c_conc/mykeAdapter.sml"

subsection{*Serialisation: semantics of SUT*}

code_printing 
  constant PUT => (SML) "!(MykeAdapter.mykeAdapter'_con ( _ ) ( _ ))"

export_code                  mykeos_interleave.test_script in SML    
  module_name TestScript file  "impl/c_conc/mykeos_test_script.sml"


end